@@ -9,7 +9,7 @@ use strict;
 use warnings;
 use POSIX ();
 use Fcntl qw/O_CREAT O_WRONLY :flock/;
-use FreezeThaw ();
+use Storable ();
 use Socket ();
 use HTTP::Response;
 use Dpkg::Arch;
@@ -111,6 +111,8 @@ sub read_config {
 							   utopic
 							   vivid
 							   wily
+							   xenial
+							   yakkety
 						      )),
 		  user => $>,
 
@@ -118,53 +120,58 @@ sub read_config {
 		  _config_file => $config_file,
 		  _path_map => {
 				'debian-changelogs' => ['packages.debian.org', 'metadata.ftp-master.debian.org'],
-				'ubuntu-changelogs' => ['changelogs.ubuntu.com']
+				'debian-appstream' => ['appstream.debian.org'],
+				'ubuntu-changelogs' => ['changelogs.ubuntu.com'],
+				'ubuntu-appstream' => ['appstream.ubuntu.com']
 			       },
 
 		  # Regexps
-		  checksum_files_regexp => '^(?:' . join('|',
-							 qw((?:Sources|Packages)(?:\.gz|\.bz2)?
+		  checksum_files_regexp => '(?:^|/)(?:' . join('|',
+							 qw((?:Sources|Packages)(?:\.(?:x|g)z|\.bz2)?
 							    (?:In)?Release
 							    Index(?:\.bz2)?
 							  )
-							) . ')$',
-		  skip_checksum_files_regexp => '^(?:' . join('|',
+							      ) . ')$',
+		  skip_checksum_files_regexp => '(?:^|/)(?:' . join('|',
 							qw((?:In)?Release
 							   Release\.gpg
 							 )
-						       ) . ')$',
-		  index_files_regexp => '^(?:' . join('|',
+								   ) . ')$',
+		  index_files_regexp => '(?:^|/)(?:' . join('|',
 						      qw(Index(?:\.bz2)?
-							 (?:Sources|Packages|release)(?:\.gz|\.bz2)?
+							 (?:Sources|Packages|release)(?:\.(?:x|g)z|\.bz2)?
 							 Release(?:\.gpg)?
 							 InRelease
 							 Contents-(?:[a-z]+-)?[a-zA-Z0-9]+\.gz
 							 (?:srclist|pkglist)\.[a-z-]+\.bz2
+							 Components-%VALID_ARCHS%\.yml\.(?:x|g)z
+							 icons-(64|128)x\g{-1}\.tar\.(?:x|g)z
 						       ),
 						      # This needs to be a separate item to avoid a warning from the
 						      # comma within qw()
 						      q(Translation-[a-z]{2,3}(?:_[A-Z]{2}(?:\.[a-zA-Z0-9-]+)?)?(?:\.gz|\.bz2|\.xz|\.lzma)?)
-						     ) . ')$',
-		  installer_files_regexp => '^(?:' . join('|',
+							   ) . ')$',
+		  installer_files_regexp => '(?:^|/)(?:' . join('|',
 							  qw(vmlinuz
 							     linux
 							     initrd\.gz
 							     (?:%VALID_PACKAGE_NAME%_%VALID_VERSION%[_\.])?changelog
 							     NEWS\.Debian
 							     %VALID_UBUNTU_RELEASE_NAMES%\.tar\.gz(?:\.gpg)?
+							     (?:by-hash/(?i:MD5SUM/[0-9a-f]{32}|SHA1/[0-9a-f]{40}|SHA256/[0-9a-f]{64}))
 							     (?:Devel|EOL)?ReleaseAnnouncement(?:\.html)?
 							     meta-release(?:-lts)?(?:-(?:development|proposed))?
 							   )
-							 ) . ')$',
+							       ) . ')$',
 		  package_files_regexp => '(?:' . join('|',
-						       qw(^%VALID_PACKAGE_NAME%_%VALID_VERSION%(?:_%VALID_ARCHS%\.(?:u|d)?deb|\.dsc|\.tar(?:\.gz|\.bz2|\.xz)|\.diff\.gz)
+						       qw((?:^|/)%VALID_PACKAGE_NAME%_%VALID_VERSION%(?:_%VALID_ARCHS%\.(?:u|d)?deb|\.dsc|\.tar(?:\.gz|\.bz2|\.xz)|\.diff\.gz)
 							  \.rpm
 							  index\.db-.+\.gz
 							  \.jigdo
 							  \.template
 							)
 						      ) .')$',
-		  pdiff_files_regexp => '^2\d{3}-\d{2}-\d{2}-\d{4}\.\d{2}\.gz$',
+		  pdiff_files_regexp => '(?:^|/)2\d{3}-\d{2}-\d{2}-\d{4}\.\d{2}\.gz$',
 		  soap_url_regexp => '^(?:http://)?bugs\.debian\.org(?::80)?/cgi-bin/soap\.cgi$',
 		 );
 
@@ -456,14 +463,26 @@ sub hashify {
     my ($href) = @_;
     return unless $$href;
     if ($$href =~ /^FrT;/) {
-	# New format: FreezeThaw
-	return (FreezeThaw::thaw($$href))[0];
-    } elsif ($$href =~ /. ./) {
+	# Deprecated format: FreezeThaw
+	if (Module::Load::Conditional::check_install(module => 'FreezeThaw')) {
+	    require FreezeThaw;
+	    return (FreezeThaw::thaw($$href))[0];
+	}
+	else {
+	    warn "Deprecated data serialisation format found. Requires libfreezethaw-perl to be installed\n";
+	}
+    }
+    elsif (Storable::read_magic $$href) {
+	# New format: Storable
+	return Storable::thaw($$href);
+    }
+    elsif ($$href =~ /. ./) {
     	# Old format: join
 	return {split(/ /, $$href)};
     } else {
-	return;
+	warn "Unrecognised serialisation method\n";
     }
+    return;
 }
 
 # Get filename from filehandle
@@ -685,18 +704,20 @@ sub extract_sums {
    my ($indexbase) = ($name =~ /([^\/]+_)(?:Index|(?:In)?Release)$/);
    $indexbase = '' unless $indexbase; # Empty by default (for Sources)
 
+   my %hash_length = (32 => 'md5', 40 => 'sha1', 64 => 'sha256');
    my ($skip,%data);
    while(<$raw>) {
        last if $AnyUncompressError;
        chomp;
-       # This flag prevents us bothering with the History section of diff_Index files
-       if (/^SHA1-(?:Current|History)/) {
-	   $skip = 1;
-       }
-       elsif (/^SHA1-Patches:/) {
+       if (/^SHA\d+-Patches:/) {
 	   $skip = 0;
        }
-       elsif (/^\s(\w{32}|\w{40}|\w{64})\s+(\d+)\s(\S+)$/) { # diff_Index/Release/Sources
+       elsif (/^SHA\d+-[a-zA-Z]+:/) {
+	   # This flag prevents us bothering with unnecessary sections
+	   # (History|Current|Download) of diff_Index files
+	   $skip = 1;
+       }
+       elsif (/^\s+([a-z0-9]{32,64})\s+(\d+)\s(\S+)$/) { # diff_Index/Release/Sources
 	   next if $skip;
 	   my $hexdigest=$1;
 	   my $size=$2;
@@ -707,19 +728,18 @@ sub extract_sums {
 	   if ($name =~ /Index$/) {
 	       $file.=".gz";
 	   }
-	   elsif ($name =~ /_Sources(?:\.gz|\.bz2)?$/) {
+	   elsif ($name =~ /_Sources(?:\.(?:x|g)z|\.bz2)?$/) {
 	       # Prepend namespace, if set
 	       $file = $namespace . $file;
 	   }
 	   $data{$file}{size} = $size;
-	   for (my $len = length($hexdigest)) { # Select algorithm based on hex length
-	       $len == 32 # md5
-		 && do { $data{$file}{md5}=$hexdigest; last; };
-	       $len == 40 # sha1
-		 && do { $data{$file}{sha1}=$hexdigest; last; };
-	       $len == 64 # sha256
-		 && do { $data{$file}{sha256}=$hexdigest; last; };
-	       warn "Unrecognised algorithm length: $len. Ignoring.";
+	   { # Select algorithm based on hex length
+	       my $len = length($hexdigest);
+	       if (exists $hash_length{$len}) {
+		   $data{$file}{$hash_length{$len}}=$hexdigest;
+	       } else {
+		   warn "Unrecognised algorithm length: $len. Ignoring.";
+	       }
 	   }
        }
        elsif(/^MD5sum:\s+([a-z0-9]{32})$/) { # Packages
@@ -738,9 +758,11 @@ sub extract_sums {
 	   # Prepend namespace, if set
 	   $data{file} = $namespace . $1;
        }
-
-       # diff_Index and Release files have no empty line at the end, so test eof() for them
-       if(!length || ($name =~ /(?:(?:In)?Release|diff_Index)$/ && $raw->eof())) { # End of record/file
+   }
+   continue {
+       # diff_Index and Release files have no empty line at the end, so also
+       # test eof() for them
+       if(!length || $raw->eof()) { # End of record/file
 	   if (exists $data{file}) {
 	       # From Packages. Convert to hash of hashes with filename as key
 	       foreach (qw(size md5 sha1 sha256)) {
@@ -750,8 +772,8 @@ sub extract_sums {
 	       delete $data{file};
 	   }
 
-	   foreach (keys %data) {
-	       $hashref->{$_} = FreezeThaw::freeze($data{$_});
+	   while (my($key,$value) = each %data) {
+	       $hashref->{$key} = Storable::freeze($value);
 	   }
 	   undef %data; # Reset
        }
@@ -841,12 +863,11 @@ sub setup_ownership {
     return;
 }
 
-# Still matches against the filename only if called with a fully qualified path
 sub is_file_type {
     my ($type,$file) = @_;
     $type .= '_files_regexp';
     die "Regexp $type not defined in config" if !exists($cfg->{$type});
-    return ((File::Spec->splitpath($file))[2] =~ $cfg->{$type});
+    return ($file =~ $cfg->{$type});
 }
 
 sub load_checksum {
